from PyQt4 import QtGui, QtCore

class SlashPleter(QtGui.QCompleter):
    def splitPath(self, path):
        return path.split('/')
    def pathFromIndex(self, idx):
        dataList=[]
        i=idx
        while i.isValid(): 
            dataList.append(self.model().data(i, self.completionRole()))
            i = i.parent()
        return '/'.join(dataList[::-1])


import re

from PyQt4.QtGui import QSyntaxHighlighter, QColor, QFont, QTextCharFormat
from PyQt4.QtCore import Qt
from .mdx_strkundr import DEL_RE, INS_RE, STRONG_RE, EMPH_RE
from .mikibook import Mikibook


class MikiHighlighter(QSyntaxHighlighter):

    WORDS = r'(?iu)[\w\']+'

    def __init__(self, parent=None):
        super(MikiHighlighter, self).__init__(parent)
        baseFontSize = 12
        NUM = 16
        self.patterns = []
        regexp = [0] * NUM
        font = [0]*NUM
        color = [0]*NUM
        color_settings = Mikibook.highlighterColors()
        # 0: html tags - <pre></pre>
        # less naive html regex
        regexp[0] = r'</?\w+((\s+\w+(\s*=\s*(?:".*?"|\'.*?\'|[^\'">\s]+))?)+\s*|\s*)/?>'
        font[0] = QFont("monospace", baseFontSize, -1)
        color[0] = QColor(color_settings[0])
        # 1: h1 - #
        regexp[1] = '^#[^#]+'
        color[1] = QColor(color_settings[1])
        font[1] = QFont("decorative", 2*baseFontSize, QFont.Bold)
        # 2: h2 - ##
        regexp[2] = '^##[^#]+'
        color[2] = QColor(color_settings[2])
        font[2] = QFont("serif", 5.0/3*baseFontSize, QFont.Bold)
        # 3: h3 - ###
        regexp[3] = '^###[^#]+'
        color[3] = QColor(color_settings[3])
        font[3] = QFont("serif", 4.0/3*baseFontSize, QFont.Bold)
        # 4: h4 and more - ####
        regexp[4] = '^####.+'
        color[4] = QColor(color_settings[4])
        font[4] = QFont("serif", baseFontSize, QFont.Bold)
        # 5: html symbols - &gt;
        regexp[5] = '&[^; ].+;'
        color[5] = QColor(color_settings[5])
        font[5] = QFont("monospace", baseFontSize, -1)
        # 6: html comments - <!-- -->
        regexp[6] = '<!--.+-->'
        color[6] = QColor(color_settings[6])
        font[6] = QFont(None, baseFontSize, -1)
        # 7: delete - ~~delete~~
        regexp[7] = DEL_RE
        color[7] = QColor(color_settings[7])
        font[7] = QFont(None, baseFontSize, -1)
        # 8: insert - __insert__
        regexp[8] = INS_RE
        font[8] = QFont(None, baseFontSize, -1)
        font[8].setUnderline(True)
        # 9: strong - **strong**
        regexp[9] = STRONG_RE
        color[9] = QColor(color_settings[9])
        font[9] = QFont(None, baseFontSize, QFont.Bold)
        # 10: emphasis - //emphasis//
        regexp[10] = EMPH_RE
        color[10] = QColor(color_settings[10])
        font[10] = QFont(None, baseFontSize, -1, True)
        # 11: links - (links) after [] or links after []:
        regexp[11] = (r'(?<=(\]\())[^\(\)]*(?=\))|'
                    '(<https?://[^>]+>)|'
                    '(<[^ >]+@[^ >]+>)')
        font[11] = QFont(None, baseFontSize, -1, True)
        font[11].setUnderline(True)
        #.setUnderlineColor("#204A87")
        # 12: link/image references - [] or ![]
        regexp[12] = r'!?\[[^\[\]]*\]'
        color[12] = QColor(color_settings[12])
        font[12] = QFont(None, baseFontSize, -1)
        # 13: blockquotes and lists -  > or - or * or 0.
        regexp[13] = r'(^>+)|(^(?:    |\t)*[0-9]+\. )|(^(?:    |\t)*- )|(^(?:    |\t)*\* )'
        color[13] = QColor(color_settings[13])
        font[13] = QFont(None, baseFontSize, -1)
        # 14: fence - ``` or ~~~
        regexp[14] = '^(?:~{3,}|`{3,}).*$'
        color[14] = QColor(color_settings[14])
        font[14] = QFont(None, baseFontSize, QFont.Bold)

        # 15: math - $$
        regexp[15] = r'^(?:\${2}).*$'
        color[15] = QColor(color_settings[15])
        font[15] = QFont(None, baseFontSize, QFont.Bold)

        for i in range(NUM):
            p = re.compile(regexp[i])
            f = QTextCharFormat()
            if font[i] != 0:
                f.setFont(font[i])
            if color[i] != 0:
                f.setForeground(color[i])
            self.patterns.append((p, f))
        self.speller = parent.speller

        fenced_font = QFont("monospace", baseFontSize, -1)
        self.fenced_block = re.compile("^(?:~{3,}|`{3,}).*$")
        self.fenced_format = QTextCharFormat()
        self.fenced_format.setFont(fenced_font)

        math_font = QFont("monospace", baseFontSize, -1)
        self.math_block = re.compile(r"^(?:\${2}).*$")
        self.math_format = QTextCharFormat()
        self.math_format.setFont(math_font)

        self.settext_h1 = re.compile('^=+$')
        self.settext_h2 = re.compile('^-+$')


    def highlightSpellcheck(self, text):
        for word_object in re.finditer(self.WORDS, str(text)):
            if not word_object.group():
                # don't bother with empty words
                continue
            if self.speller and not self.speller.check(word_object.group()):
                current_format = self.format(word_object.start())
                current_format.setUnderlineColor(Qt.red)
                current_format.setUnderlineStyle(QTextCharFormat.SpellCheckUnderline)
                self.setFormat(word_object.start(),
                    word_object.end() - word_object.start(), current_format)

    def highlightBlock(self, text):
        # highlight patterns
        for i in range(0, len(self.patterns)):
            p = self.patterns[i]
            for match in p[0].finditer(text):
                self.setFormat(
                    match.start(), match.end() - match.start(), p[1])

        if text == '' and self.currentBlock().next().text() != '':
            self.setCurrentBlockState(5)
        elif self.previousBlockState() == 3:
            self.setFormat(0, len(text), self.patterns[1][1])
            self.setCurrentBlockState(0)
        elif self.previousBlockState() == 4:
            self.setFormat(0, len(text), self.patterns[2][1])
            self.setCurrentBlockState(0)
        else:
            # escape highlights in fenced_block
            m = self.fenced_block.match(text)
            m2 = self.math_block.match(text)
            self.setCurrentBlockState(0)
            if self.previousBlockState() not in (1,2):
                if m:
                    self.setCurrentBlockState(1)
                elif m2:
                    self.setCurrentBlockState(2)
            elif self.previousBlockState() == 1:
                if m:
                    self.setCurrentBlockState(0)
                else:
                    self.setCurrentBlockState(1)
                    self.setFormat(0, len(text), self.fenced_format)
            elif self.previousBlockState() == 2:
                if m2:
                    self.setCurrentBlockState(0)
                else:
                    self.setCurrentBlockState(2)
                    self.setFormat(0, len(text), self.math_format)
            else:
                if self.settext_h1.match(self.currentBlock().next().text()) and text != '':
                    self.setFormat(0, len(text), self.patterns[1][1])
                    self.setCurrentBlockState(3)
                elif self.settext_h2.match(self.currentBlock().next().text()) and text != '':
                    self.setFormat(0, len(text), self.patterns[2][1])
                    self.setCurrentBlockState(4)
        self.highlightSpellcheck(text)




import argparse
import os
import re
import sys
import signal

from PyQt4.QtCore import QSettings
from PyQt4.QtGui import QApplication, QIcon, QMessageBox

import mikidown.mikidown_rc
from .config import Setting
from .generator import Generator
from .mikitray import MikiTray
from .mikiwindow import MikiWindow
from .mikibook import Mikibook
from .sandbox import Sandbox

sys.path.append(os.path.dirname(__file__))

# http://code.activestate.com/recipes/578453-python-single-instance-cross-platform/

def main():

    parser = argparse.ArgumentParser(description='A note taking application, featuring markdown syntax')
    subparsers = parser.add_subparsers(dest='command')
    parser_generate = subparsers.add_parser('generate',
        help='generate a static html site from notebook')
    parser_preview = subparsers.add_parser('preview',
        help='automatically regenerate html site when notes modified')
    parser_preview.add_argument('-p', '--port', dest='port',
        type=int, help='port number')
    parser_sandbox = subparsers.add_parser('sandbox',
        help='for test purpose, all notes will be lost when exit')
    args = parser.parse_args()

    if args.command == 'generate':
        generator = Generator(os.getcwd())
        generator.generate()
        sys.exit(0)
    elif args.command == 'sandbox':
        app = QApplication(sys.argv)
        sandbox = Sandbox()
        app.aboutToQuit.connect(sandbox.cleanUp)
        sys.exit(app.exec_())
    elif args.command == 'preview':
        generator = Generator(os.getcwd())
        if args.port:
            generator.preview(args.port)
        else:
            generator.preview()


    # Instantiate a QApplication first.
    # Otherwise, Mikibook.create() won't function.
    app = QApplication(sys.argv)
    print(sys.argv)

    # Read notebookList, open the first notebook.
    notebooks = Mikibook.read()
    if len(notebooks) == 0:
        Mikibook.create()
        notebooks = Mikibook.read()

    if len(notebooks) != 0:
        #"""
        if os.path.exists(Mikibook.lockpath) and args.command != 'index':
            ret = QMessageBox.question(None, "mikidown - lock file exists", ("It looks like the lock file for "
                "mikidown already exists. Is mikidown currently running? "
                "Click no to remove the lock file before rerunning mikidown."), buttons=QMessageBox.Yes|QMessageBox.No)
            if ret == QMessageBox.Yes:
                sys.exit(1)
            else:
                os.remove(Mikibook.lockpath)
                sys.exit(0)
            exit_code = app.exec_()
        else:
            print("Applying single instance per user lock.")
            lock_fh = os.open(Mikibook.lockpath, os.O_CREAT | os.O_EXCL | os.O_RDWR)
        #"""
        settings = Setting(notebooks)
        # Initialize application and main window.
        icon = QIcon(":/icons/mikidown.svg")
        app.setWindowIcon(icon)
        window = MikiWindow(settings)
        window.show()
        window.restore()        # Restore after window show.
        tray = MikiTray(icon, window)
        tray.show()

        #"""
        def cleanup(signum, frame):
            #we need to do this to remove the lock at the end
            app.closeAllWindows()

        if args.command:
            signal.signal(signal.SIGTERM, cleanup)
        #"""
        exit_code = app.exec_()

        #"""
        print("Removing single instance per user lock.")
        if os.path.exists(Mikibook.lockpath) and args.command != 'index':
            os.close(lock_fh)
            os.remove(Mikibook.lockpath)
        #"""
        sys.exit(exit_code)

if __name__ == '__main__':
    main()



import os
import re
from PyQt4.QtCore import QDir, QFile, QSettings
from whoosh import fields
import markdown

NOT_EXT = re.compile(r"Failed to initiate extension '([^']+)': 'module' object has no attribute 'makeExtension'")

__appname__ = 'mikidown'
__version__ = '0.3.7' # we should really change this to a tuple

class Setting():

    def __init__(self, notebooks):

        # Index directory of whoosh, located in notebookPath.
        self.schema = fields.Schema(
            path = fields.TEXT(stored=True),
            title = fields.TEXT(stored=True),
            content = fields.TEXT(stored=True),
            tags = fields.KEYWORD(commas=True))

        self.notebookName = notebooks[0][0]
        self.notebookPath = notebooks[0][1]
        self.notePath = os.path.join(self.notebookPath, "notes").replace(os.sep, '/')
        self.htmlPath = os.path.join(self.notebookPath, "html", "notes").replace(os.sep, '/')
        self.indexdir = os.path.join(self.notePath, ".indexdir").replace(os.sep, '/')
        self.attachmentPath = os.path.join(self.notebookPath, "attachments").replace(os.sep, '/')
        self.configfile = os.path.join(self.notebookPath, "notebook.conf").replace(os.sep, '/')
        cssPath = os.path.join(self.notebookPath, "css").replace(os.sep, '/')
        self.cssfile = os.path.join(cssPath, "notebook.css").replace(os.sep, '/')
        self.searchcssfile = os.path.join(cssPath, "search-window.css").replace(os.sep, '/')
        self.qsettings = QSettings(self.configfile, QSettings.IniFormat)

        if os.path.exists(self.configfile):
            self.extensions = readListFromSettings(self.qsettings,
                                                   "extensions")
            self.fileExt = self.qsettings.value("fileExt")
            self.attachmentImage = self.qsettings.value("attachmentImage")
            self.attachmentDocument = self.qsettings.value("attachmentDocument")
            self.version = self.qsettings.value("version")
            self.geometry = self.qsettings.value("geometry")
            self.windowstate = self.qsettings.value("windowstate")
            self.mathjax = self.qsettings.value('mathJax')
            if 'extensionsConfig' not in set(self.qsettings.childGroups()):
                self.extcfg = self.qsettings.value('extensionsConfig',  defaultValue={})
                writeDictToSettings(self.qsettings, 'extensionsConfig', self.extcfg)
            else:
                self.extcfg = readDictFromSettings(self.qsettings, 'extensionsConfig')
        else:
            self.extensions = []
            self.fileExt = ""
            self.attachmentImage = []
            self.attachmentDocument = []
            self.version = None
            self.geometry = None
            self.windowstate = None
            self.mathjax = ''
            self.extcfg = {}

        self.faulty_exts=[]

        # Default enabled python-markdown extensions.
        # http://pythonhosted.org/Markdown/extensions/index.html
        if not self.extensions:
            self.extensions = [
                   'nl2br'           # newline to break
                 , 'strkundr'        # bold-italics-underline-delete style
                 , 'codehilite'      # code syntax highlight
                 , 'fenced_code'     # code block
                 , 'headerid'        # add id to headers
                 , 'headerlink'      # add anchor to headers
                 , 'footnotes'
                 , 'mdx_asciimathml'
                 ]
            writeListToSettings(self.qsettings, "extensions", self.extensions)

        while True:
             print(self.extensions)
             try:
                 markdown.markdown("",extensions=self.extensions)
             except AttributeError as e:
                 remove_this = NOT_EXT.findall(e.args[0])[0]
                 if remove_this in self.extensions:
                     print("Found invalid markdown extension",remove_this,". Please consider removing it.")
                     print('If you want to permanently disable this, just hit OK in the Notebook Settings dialog')
                     self.extensions.remove(remove_this)
                     self.faulty_exts.append(remove_this)
             except ImportError as e:
                 if e.name.startswith('mdx_') and e.name[4:] in self.extensions:
                     print('Found missing markdown extension', e.name[4:], ', temporarily disabling.')
                     print('If you want to permanently disable this, just hit OK in the Notebook Settings dialog')
                     self.extensions.remove(e.name[4:])
                     self.faulty_exts.append(e.name[4:])
                 elif e.name in self.extensions:
                     print('Found missing markdown extension', e.name, ', temporarily disabling.')
                     print('If you want to permanently disable this, just hit OK in the Notebook Settings dialog')
                     self.extensions.remove(e.name)
                     self.faulty_exts.append(e.name)
             else:
                 self.md = markdown.Markdown(self.extensions, extension_configs=self.extcfg)
                 break

        # Default file extension name
        if not self.fileExt:
            self.fileExt = ".md"
            self.qsettings.setValue("fileExt", self.fileExt)

        # Image file types that will be copied to attachmentDir
        # Inserted as image link
        if not self.attachmentImage:
            self.attachmentImage = [".jpg", ".jpeg", ".png", ".gif", ".svg"]
            self.qsettings.setValue("attachmentImage", self.attachmentImage)

        # Document file types that will be copied to attachmentDir
        # Inserted as link
        if not self.attachmentDocument:
            self.attachmentDocument = [".pdf", ".doc", ".odt"]
            self.qsettings.setValue("attachmentDocument", self.attachmentDocument)

        # Migrate notebookPath to v0.3.0 folder structure
        if not self.version:
            notebookDir = QDir(self.notebookPath)

            # move all markdown files to notes/
            dirList = notebookDir.entryList(QDir.Dirs | QDir.NoDotAndDotDot)
            if 'css' in dirList:
                dirList.remove('css')
            fileList = notebookDir.entryList(['*.md', '*.mkd', '*.markdown'])
            notebookDir.mkdir('notes')
            for d in dirList + fileList:
                notebookDir.rename(d, os.path.join('notes', d).replace(os.sep, '/'))

            # remove .indexdir folder
            oldIndexDir = QDir(os.path.join(self.notebookPath, '.indexdir'.replace(os.sep, '/')))
            indexFileList = oldIndexDir.entryList()
            for f in indexFileList:
                oldIndexDir.remove(f)
            notebookDir.rmdir('.indexdir')

            # rename notes.css to css/notebook.css
            oldCssFile = os.path.join(self.notebookPath, 'notes.css').replace(os.sep, '/')
            QDir().mkpath(cssPath)
            if os.path.exists(oldCssFile):
                QFile.rename(oldCssFile, self.cssfile)

            self.version = '0'

        if not self.mathjax:
            self.mathjax = 'http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML'
            self.qsettings.setValue('mathJax', self.mathjax)

    def saveGeometry(self, geometry):
        self.qsettings.setValue("geometry", geometry)

    def saveWindowState(self, state):
        self.qsettings.setValue("windowstate", state)

    def recentViewedNotes(self):
        return readListFromSettings(self.qsettings, "recentViewedNoteList")

    def updateRecentViewedNotes(self, notesList):
        writeListToSettings(self.qsettings, "recentViewedNoteList", notesList)

def readListFromSettings(settings, key):
    if not settings.contains(key):
        return []
    value = settings.value(key)
    if isinstance(value, str):
        return [value]
    else:
        return value


def writeListToSettings(settings, key, value):
    if len(value) >= 1:
        settings.setValue(key, value)
    else:
        settings.remove(key)

def readDictFromSettings(settings, key):
    data={}
    settings.beginGroup(key)
    for k in settings.childGroups():
        settings.beginGroup(k)
        key_data = []
        for k2 in settings.childKeys():
            key_data.append((k2, settings.value(k2)))
        settings.endGroup()
        data[k]=key_data
    settings.endGroup()
    return data

def writeDictToSettings(settings, key, value):
    settings.beginGroup(key)
    for k in value.keys():
        settings.beginGroup(k)
        for v in value[k]:
            settings.setValue(v[0], v[1])
        settings.endGroup()
    settings.endGroup()



"""
Notebook management module.
"""

import os
import markdown
from copy import deepcopy
from PyQt4.QtCore import Qt, QDir, QFile, QSettings, QSize
from PyQt4.QtGui import (QAbstractItemDelegate, QAbstractItemView, QColor, QDialog, QDialogButtonBox, 
                         QFileDialog, QFont, QGridLayout, QLabel, QLineEdit, QListWidget, QListWidgetItem,
                         QPen, QPushButton, QStyle, QVBoxLayout, QTabWidget, QWidget, QBrush, QTreeWidget,
                         QTreeWidgetItem, QSpinBox, QScrollArea)

import mikidown
try:
    import slickpicker
    BETTER_COLOR_PICKER=True
except ImportError as e:
    print("Can't find slickpicker, falling back to QLineEdit for editing mikidown colors")
    BETTER_COLOR_PICKER=False
from .utils import allMDExtensions
from .config import Setting, readListFromSettings, writeListToSettings, writeDictToSettings


class ListDelegate(QAbstractItemDelegate):
    """ Customize view and behavior of notebook list """

    def __init__(self, parent=None):
        super(ListDelegate, self).__init__(parent)

    def paint(self, painter, option, index):
        r = option.rect

        if option.state & QStyle.State_Selected:
            painter.fillRect(r, self.parent().palette().highlight())
            fontPen = QPen(self.parent().palette().highlightedText(), 1, Qt.SolidLine)
        else:
            painter.fillRect(r, self.parent().palette().base())
            fontPen = QPen(self.parent().palette().text(), 1, Qt.SolidLine)

        painter.setPen(fontPen)

        name = index.data(Qt.DisplayRole)
        path = index.data(Qt.UserRole)

        imageSpace = 10
        # notebook name
        r = option.rect.adjusted(imageSpace, 0, -10, -20)
        name_font = QFont(self.parent().font())
        name_font.setPointSize(10)
        name_font.setBold(True)
        painter.setFont(name_font)
        painter.drawText(r.left(), r.top(), r.width(), r.height(), 
                         Qt.AlignBottom|Qt.AlignLeft, name)
        # notebook path
        path_font = QFont(self.parent().font())
        path_font.setPointSize(8)
        r = option.rect.adjusted(imageSpace, 20, -10, 0)
        painter.setFont(path_font)
        painter.drawText(r.left(), r.top(), r.width(), r.height(), 
                         Qt.AlignLeft, path)

    def sizeHint(self, option, index):
        return QSize(200, 40)

class NotebookExtSettingsDialog(QDialog):
    def __init__(self, parent=None, cfg_list=[]):
        super(NotebookExtSettingsDialog, self).__init__(parent)
        self.extCfgEdit = QTreeWidget()
        self.extCfgEdit.setHeaderLabels(['Property', 'Value'])
        self.addRow = QPushButton('+')
        self.removeRow = QPushButton('-')
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok |
                                          QDialogButtonBox.Cancel)

        layout = QGridLayout(self)
        layout.addWidget(self.extCfgEdit,0,0,1,2)
        layout.addWidget(self.addRow,1,0,1,1)
        layout.addWidget(self.removeRow,1,1,1,1)
        layout.addWidget(self.buttonBox,2,0,1,2)
        self.initCfgPanel(cfg_list)

        self.addRow.clicked.connect(self.actionAdd)
        self.removeRow.clicked.connect(self.actionRemove)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

    def initCfgPanel(self, cfg_list):
        for item in cfg_list:
            self.actionAdd(prop_name=item[0], prop_val=item[1])

    def actionRemove(self):
        item = self.extCfgEdit.currentItem()
        row = self.extCfgEdit.indexOfTopLevelItem(item)
        self.extCfgEdit.takeTopLevelItem(row)

    def actionAdd(self, checked=False, prop_name='', prop_val=''):
        item = QTreeWidgetItem(self.extCfgEdit, [prop_name, prop_val])
        item.setFlags(item.flags()|Qt.ItemIsEditable)
        #self.extCfgEdit.addTopLevelItem(item)

    def configToList(self):
        items = []
        for i in range(self.extCfgEdit.topLevelItemCount()):
            witem = self.extCfgEdit.topLevelItem(i)
            items.append((witem.text(0), witem.text(1)))
        return items

class NotebookSettingsDialog(QDialog):
    """GUI for adjusting notebook settings"""
    def __init__(self, parent=None):
        super(NotebookSettingsDialog, self).__init__(parent)
        #widgets for tab 1
        self.mdExts = QListWidget()
        self.mjEdit = QLineEdit()
        self.moveUp = QPushButton('<<')
        self.moveDown = QPushButton('>>')
        self.configureExtension = QPushButton('Edit Settings for this extension')
        self.tmpdict = deepcopy(self.parent().settings.extcfg)
        
        #widgets for tab 2
        self.fExtEdit = QLineEdit()
        self.attImgEdit = QLineEdit()
        self.attDocEdit = QLineEdit()
        # mandatory button box
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok |
                                          QDialogButtonBox.Cancel)
        
        #tab panels
        tabs = QTabWidget()
        markupTab = QWidget()
        fileExtsTab = QWidget()
        tabs.addTab(markupTab, "Markdown")
        tabs.addTab(fileExtsTab, "File extensions")
        
        #initialization functions
        self.initExtList()
        self.mdExts.setDragDropMode(QAbstractItemView.InternalMove)
        self.mjEdit.setText(self.parent().settings.mathjax)
        self.attImgEdit.setText(', '.join(self.parent().settings.attachmentImage))
        self.attDocEdit.setText(', '.join(self.parent().settings.attachmentDocument))
        self.fExtEdit.setText(self.parent().settings.fileExt)
        
        #set up tab 1
        layout=QGridLayout(markupTab)
        layout.addWidget(QLabel("Markdown extensions"),0,0,1,4)
        layout.addWidget(self.mdExts,1,0,1,4)
        layout.addWidget(self.moveUp,2,0,1,1)
        layout.addWidget(self.moveDown,2,1,1,1)
        layout.addWidget(self.configureExtension,2,2,1,2)
        layout.addWidget(QLabel("MathJax Location"),3,0,1,1)
        layout.addWidget(self.mjEdit,3,1,1,3)
        
        #set up tab 2
        layout=QGridLayout(fileExtsTab)
        layout.addWidget(QLabel("Note file extension"),0,0,1,1)
        layout.addWidget(QLabel("Image file extension"),1,0,1,1)
        layout.addWidget(QLabel("Document file extension"),2,0,1,1)
        layout.addWidget(self.fExtEdit,0,1,1,1)
        layout.addWidget(self.attImgEdit,1,1,1,1)
        layout.addWidget(self.attDocEdit,2,1,1,1)
        
        #put it together
        vlayout = QVBoxLayout(self)
        vlayout.addWidget(tabs)
        vlayout.addWidget(self.buttonBox)

        #setup signal handlers
        self.moveUp.clicked.connect(self.moveItemUp)
        self.configureExtension.clicked.connect(self.configExt)
        self.moveDown.clicked.connect(self.moveItemDown)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

    def configExt(self, checked=False, ext=None):
        if ext is None:
            ext = self.mdExts.currentItem().text()
        cfg = self.tmpdict.get(ext,[])
        dialog = NotebookExtSettingsDialog(cfg_list=cfg)
        done = dialog.exec()
        if done:
            self.tmpdict[ext] = dialog.configToList()

    def initExtList(self):
        extset=set(self.parent().settings.extensions)
        #for easier performance in checking
        for ext in self.parent().settings.extensions:
            item = QListWidgetItem(ext, self.mdExts)
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setCheckState(Qt.Checked)

        for ext in self.parent().settings.faulty_exts:
            item = QListWidgetItem(ext, self.mdExts)
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setBackground(QBrush(QColor('red')))
            item.setForeground(QBrush(QColor('black')))
            item.setCheckState(Qt.Checked)

        for ext in allMDExtensions():
            if ext in extset: continue
            item = QListWidgetItem(ext, self.mdExts)
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setCheckState(Qt.Unchecked)
            #self.mdExts.addItem(item)

    def moveItemUp(self):
        item = self.mdExts.currentItem()
        row = self.mdExts.currentRow()
        if row != 0:
            # self.mdExts.removeItemWidget(item)
            self.mdExts.takeItem(row)
            self.mdExts.insertItem(row-1, item)
            self.mdExts.setCurrentRow(row-1)

    def moveItemDown(self):
        item = self.mdExts.currentItem()
        row = self.mdExts.currentRow()
        count = self.mdExts.count()
        if row != count-1:
            self.mdExts.takeItem(row)
            self.mdExts.insertItem(row+1, item)
            self.mdExts.setCurrentRow(row+1)


    def accept(self):
        #write to settings first
        msettings = self.parent().settings
        nbsettings = msettings.qsettings
        
        nbsettings.setValue('mathJax', self.mjEdit.text())
        extlist = []
        for i in range(self.mdExts.count()):
            item = self.mdExts.item(i)
            if item.checkState() == Qt.Checked:
                extlist.append(item.text())
        writeListToSettings(nbsettings, 'extensions', extlist)
        writeListToSettings(nbsettings, 'attachmentImage', self.attImgEdit.text().split(", "))
        writeListToSettings(nbsettings, 'attachmentDocument', self.attDocEdit.text().split(", "))
        writeDictToSettings(nbsettings, 'extensionsConfig', self.tmpdict)
        
        #then to memory
        msettings.extensions = extlist
        msettings.mathjax = self.mjEdit.text()
        msettings.attachmentDocument = readListFromSettings(nbsettings, 'attachmentDocument')
        msettings.attachmentImage = readListFromSettings(nbsettings, 'attachmentImage')
        msettings.extcfg.update(self.tmpdict)
        msettings.md = markdown.Markdown(msettings.extensions, extension_configs=msettings.extcfg)
        
        #then make mikidown use these settings NOW
        curitem=self.parent().notesTree.currentItem()
        self.parent().currentItemChangedWrapper(curitem, curitem)
        QDialog.accept(self)

class NotebookListDialog(QDialog):
    """ Functions to display, create, remove, modify notebookList """

    def __init__(self, parent=None):
        super(NotebookListDialog, self).__init__(parent)

        self.notebookList = QListWidget()
        self.moveUp = QPushButton('<<')
        self.moveDown = QPushButton('>>')
        self.add = QPushButton('Add')
        self.remove = QPushButton('Remove')
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok |
                                          QDialogButtonBox.Cancel)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        layout = QGridLayout()
        layout.addWidget(self.notebookList, 0, 0, 4, 6)
        layout.addWidget(self.moveUp, 1, 6)
        layout.addWidget(self.moveDown, 2, 6)
        layout.addWidget(self.add, 4, 0)
        layout.addWidget(self.remove, 4, 1)
        layout.addWidget(self.buttonBox, 4, 5, 1, 2)
        self.setLayout(layout)

        self.notebookList.setItemDelegate(ListDelegate(self.notebookList))

        self.notebookList.currentRowChanged.connect(self.updateUi)
        self.add.clicked.connect(self.actionAdd)
        self.remove.clicked.connect(self.actionRemove)
        self.moveUp.clicked.connect(self.moveItemUp)
        self.moveDown.clicked.connect(self.moveItemDown)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.initList()

    def initList(self):
        self.notebookList.clear()
        notebooks = Mikibook.read()
        for nb in notebooks:
            item = QListWidgetItem()
            item.setData(Qt.DisplayRole, nb[0])
            item.setData(Qt.UserRole, nb[1])
            self.notebookList.addItem(item)

        self.updateUi(len(notebooks) != 0)
        self.notebookList.setCurrentRow(0)
            # QListWidgetItem(nb, self.notebookList)

    def updateUi(self, row):
        flag = (row != -1)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(flag)
        self.remove.setEnabled(flag)
        self.moveUp.setEnabled(flag)
        self.moveDown.setEnabled(flag)

    def actionAdd(self):
        Mikibook.create()
        self.initList()
        count = self.notebookList.count()
        self.notebookList.setCurrentRow(count-1)

    def actionRemove(self):
        item = self.notebookList.currentItem()
        row = self.notebookList.currentRow()
        name = item.data(Qt.DisplayRole)
        path = item.data(Qt.UserRole)
        self.notebookList.takeItem(row)

        Mikibook.remove(name, path)

    def moveItemUp(self):
        item = self.notebookList.currentItem()
        row = self.notebookList.currentRow()
        if row != 0:
            # self.notebookList.removeItemWidget(item)
            self.notebookList.takeItem(row)
            self.notebookList.insertItem(row-1, item)
            self.notebookList.setCurrentRow(row-1)

    def moveItemDown(self):
        item = self.notebookList.currentItem()
        row = self.notebookList.currentRow()
        count = self.notebookList.count()
        if row != count-1:
            self.notebookList.takeItem(row)
            self.notebookList.insertItem(row+1, item)
            self.notebookList.setCurrentRow(row+1)

    def accept(self):
        notebookPath = self.notebookList.currentItem().data(Qt.UserRole)
        notebookName = self.notebookList.currentItem().data(Qt.DisplayRole)
        settings = Setting([[notebookName, notebookPath]])
        window = mikidown.MikiWindow(settings)
        window.show()
        count = self.notebookList.count()
        notebooks = []
        for i in range(count):
            name = self.notebookList.item(i).data(Qt.DisplayRole)
            path = self.notebookList.item(i).data(Qt.UserRole)
            notebooks.append([name, path])
            Mikibook.write(notebooks)

        QDialog.accept(self)


class NewNotebookDlg(QDialog):
    def __init__(self, parent=None):
        super(NewNotebookDlg, self).__init__(parent)
        self.setWindowTitle('Add Notebook - mikidown')
        tipLabel = QLabel('Choose a name and folder for your notebook.' +
                          '\nThe folder can be an existing notebook folder.')
        self.nameEditor = QLineEdit()
        self.nameEditor.setText('Notes')
        nameLabel = QLabel('Name:')
        nameLabel.setBuddy(self.nameEditor)
        self.pathEditor = QLineEdit()
        # self.pathEditor.setText('~/mikidown')
        self.pathEditor.setText(os.path.expanduser('~').replace(os.sep,'/')+'/mikinotes')
        pathLabel = QLabel('Path:')
        pathLabel.setBuddy(self.pathEditor)
        browse = QPushButton('Browse')
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok |
                                     QDialogButtonBox.Cancel)

        grid = QGridLayout()
        grid.setRowMinimumHeight(1, 10)
        grid.setRowMinimumHeight(4, 10)
        grid.addWidget(tipLabel, 0, 0, 1, 4)
        grid.addWidget(nameLabel, 2, 0)
        grid.addWidget(self.nameEditor, 2, 1, 1, 4)
        grid.addWidget(pathLabel, 3, 0)
        grid.addWidget(self.pathEditor, 3, 1, 1, 4)
        grid.addWidget(browse, 3, 5)
        grid.addWidget(buttonBox, 5, 4, 1, 2)
        self.setLayout(grid)

        browse.clicked.connect(self.browse)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

    def browse(self):
        default = os.path.expanduser('~')
        path = QFileDialog.getExistingDirectory(self,
                                                "Select Folder",
                                                default,
                                                QFileDialog.ShowDirsOnly)
        self.pathEditor.setText(path)

    def closeEvent(self, event):
        event.accept()

class MikidownHighlightCfgWidget(QWidget):
    def __init__(self, parent=None):
        super(MikidownHighlightCfgWidget, self).__init__(parent)
        layout = QGridLayout(self)
        colors = Mikibook.highlighterColors()
        for i in range(16):
            layout.addWidget(QLabel(Mikibook.highlighter_labels[i]),i,0,1,1)
            if BETTER_COLOR_PICKER:
                layout.addWidget(slickpicker.QColorEdit(colors[i]),i,1,1,1)
            else:
                layout.addWidget(QLineEdit(colors[i]),i,1,1,1)

    def configToList(self):
        items=[]
        for i in range(16):
            if BETTER_COLOR_PICKER:
                items.append(self.layout().itemAtPosition(i,1).widget().lineEdit.text())
            else:
                items.append(self.layout().itemAtPosition(i,1).widget().text())
        return items

class MikidownCfgDialog(QDialog):
    def __init__(self, parent=None):
        super(MikidownCfgDialog, self).__init__(parent)
        #tab = QWidget()
        #tab2 = QWidget()
        self.recentNotesCount = QSpinBox()
        recent_notes_n = Mikibook.settings.value('recentNotesNumber',type=int, defaultValue=20)
        self.recentNotesCount.setValue(recent_notes_n)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok |
                                          QDialogButtonBox.Cancel)
        self.hltCfg = MikidownHighlightCfgWidget(parent=self)

        layout = QGridLayout(self)
        layout.addWidget(QLabel("# of recently viewed notes to keep"),0,0,1,1)
        layout.addWidget(self.recentNotesCount,0,1,1,1)
        qs = QScrollArea(self)
        qs.setWidget(self.hltCfg)
        layout.addWidget(qs,1,0,1,2)
        layout.addWidget(self.buttonBox,2,0,1,2)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

    def accept(self):
        Mikibook.settings.setValue('recentNotesNumber', self.recentNotesCount.value())
        Mikibook.setHighlighterColors(self.hltCfg.configToList())

        #then make mikidown use these settings NOW
        self.parent().loadHighlighter()
        QDialog.accept(self)

class Mikibook():
    # ~/.config/mikidown/mikidown.conf
    settings = QSettings(QSettings.IniFormat, QSettings.UserScope, 'mikidown', 'mikidown')
    lockpath = os.path.join(os.path.dirname(settings.fileName()),'lock').replace(os.sep,'/')
    highlighter_labels = ['HTML Tags', '1<sup>st</sup> LVL headers', '2<sup>nd</sup> LVL headers',
                         '3<sup>rd</sup> LVL headers', '4<sup>th</sup> and lower LVL headers',
                         'HTML Symbols','HTML comments','Strikethrough','Underline','Bold', 'Italics',
                         'Links', 'Links and images', 'Block Quotes','Fenced Code','Math']

    @staticmethod
    def highlighterColors():
        items = []
        defaults = [ '#A40000',
                    "#4E9A06",
                    "#4E9A06",
                    "#4E9A06",
                    "#4E9A06",
                    "#A40000",
                    "#ff0037", #italic, not used
                    "#888A85",
                    "#888A85",
                    "#F57900",
                    "#F57900",
                    "#204A87", #underline color, not used atm
                    "#204A87",
                    "#F57900",
                    "#F57900",
                    "#F5006E"]
        size = Mikibook.settings.beginReadArray('highlighting')
        if size == 0:
            Mikibook.settings.endArray()
            Mikibook.setHighlighterColors(defaults)
            size = Mikibook.settings.beginReadArray('highlighting')
        for i in range(16):
            Mikibook.settings.setArrayIndex(i)
            items.append(Mikibook.settings.value('color', defaultValue=defaults[i], type=str))
        Mikibook.settings.endArray()
        return items

    @staticmethod
    def setHighlighterColors(items):
        Mikibook.settings.beginWriteArray('highlighting')
        for i,val in enumerate(items):
            Mikibook.settings.setArrayIndex(i)
            Mikibook.settings.setValue('color', val)
        Mikibook.settings.endArray()

    @staticmethod
    def read():
        """ Read notebook list from config file """
        version = Mikibook.settings.value("version", defaultValue=None)
        if not version: #before 0.3.4, since we're migrating the notebooklist to be plaintext
            Mikibook.nbListMigration()
            Mikibook.settings.setValue("version", "0") #dummy value until mikiwindow properly sets this
        items = []
        size = Mikibook.settings.beginReadArray("notebookList")
        for i in range(size):
            Mikibook.settings.setArrayIndex(i)
            items.append((Mikibook.settings.value('name', type=str),
                        Mikibook.settings.value('path', type=str)))
        Mikibook.settings.endArray()
        return items

    @staticmethod
    def nbListMigration():
        #print("nbListMigration")
        books = readListFromSettings(Mikibook.settings, 'notebookList')
        Mikibook.write(books)

    @staticmethod
    def write(notebooks):
        """ Write notebook list to config file """ 
        Mikibook.settings.beginWriteArray("notebookList")
        for i, val in enumerate(notebooks):
            Mikibook.settings.setArrayIndex(i)
            Mikibook.settings.setValue('name', val[0])
            Mikibook.settings.setValue('path', val[1])
        Mikibook.settings.endArray()
        #print("Mikibook.write:", Mikibook.settings.value("notebookList/size"))

    @staticmethod
    def create():
        """ Display a dialog to set notebookName and notebookPath """
        newNotebook = NewNotebookDlg()
        if newNotebook.exec_():
            notebookName = newNotebook.nameEditor.text()
            notebookPath = newNotebook.pathEditor.text()
            Mikibook.initialise(notebookName, notebookPath)

            notebooks = Mikibook.read()
            #print("Mikibook.create -> .read:",notebooks)
            notebooks.append([notebookName, notebookPath])
            Mikibook.write(notebooks)
            #print("Mikibook.create -> .read(2):", Mikibook.read())

    @staticmethod
    def initialise(notebookName, notebookPath):
        """ Called by create()
        A notebook directory will be initialised to:
            css/  notebook.conf  notes/
        """

        # QDir().mkpath will create all necessary parent directories
        QDir().mkpath(os.path.join(notebookPath, "notes").replace(os.sep,'/'))
        QDir().mkpath(os.path.join(notebookPath, "css").replace(os.sep,'/'))
        cssFile = os.path.join(notebookPath, "css", "notebook.css").replace(os.sep,'/')
        searchCssFile = os.path.join(notebookPath, "css", "search-window.css").replace(os.sep,'/')
        cssTemplate = "/usr/share/mikidown/notebook.css"
        searchCssTemplate = "/usr/share/mikidown/search-window.css"
        if not os.path.exists(cssTemplate):
            cssTemplate = os.path.join(
                os.path.dirname(__file__), "css", "sphinx.css").replace(os.sep,'/')
        if not os.path.exists(searchCssTemplate):
            searchCssTemplate = os.path.join(
                os.path.dirname(__file__), "css" , "search-window.css").replace(os.sep,'/')
        # If //cssFile// already exists, copy() returns false!
        print(cssTemplate)
        print(searchCssTemplate)
        QFile.copy(cssTemplate, cssFile)
        QFile.copy(searchCssTemplate, searchCssFile)

    @staticmethod
    def remove(name, path):
        notebooks = Mikibook.read()
        notebooks.remove([name, path])
        Mikibook.write(notebooks)



from PyQt4.QtCore import *
from PyQt4.QtGui import *
import re

class FindReplaceDialog(QDialog):
  def __init__(self, parent = None):
    super(FindReplaceDialog, self).__init__(parent)
    grid = QGridLayout(self)
    
    self.searchInput = QLineEdit(self)
    self.replaceInput = QLineEdit(self)
    
    self.matchCase = QCheckBox(self.tr('Match case'), self)
    self.wholeWords = QCheckBox(self.tr('Whole words'), self)
    self.reCheck = QCheckBox(self.tr('Regex'), self)

    self.nextButton = QPushButton(self.tr("Next"),self)
    self.prevButton = QPushButton(self.tr("Previous"),self)
    self.replaceButton = QPushButton(self.tr("Replace"),self)
    self.replaceAllButton = QPushButton(self.tr("Replace All"),self)
    
    self.nextButton.clicked.connect(self.find)
    self.prevButton.clicked.connect(lambda: self.find(back = True))
    self.replaceButton.clicked.connect(self.replace)
    self.replaceAllButton.clicked.connect(self.replaceAll)
    self.searchInput.returnPressed.connect(self.find)
    self.replaceInput.returnPressed.connect(self.replace)
    
    grid.addWidget(QLabel(self.tr("Search")),0,0)
    grid.addWidget(QLabel(self.tr("    Options")),1,0)
    grid.addWidget(QLabel(self.tr("Replace")),2,0)
    
    grid.addWidget(self.searchInput,0,1)
    grid.addWidget(self.nextButton,0,2)
    grid.addWidget(self.prevButton,0,3)
    
    grid.addWidget(self.matchCase,1,1)
    grid.addWidget(self.wholeWords,1,2)
    grid.addWidget(self.reCheck,1,3)
    
    grid.addWidget(self.replaceInput,2,1)
    grid.addWidget(self.replaceButton,2,2)
    grid.addWidget(self.replaceAllButton,2,3)
  
  def replace(self,autofind = True):
    if autofind:
      self.find()
    contents = self.parent().textCursor().selectedText()

    if self.reCheck.isChecked():
      search_term = self.searchInput.text()
    else:
      search_term = re.escape(self.searchInput.text())

    replace_term = self.replaceInput.text()
    flags = 0
    if not self.matchCase.isChecked():
      flags = re.IGNORECASE
    if self.wholeWords.isChecked():
      search_term = r'\b{}\b'.format(search_term)

    contents = re.sub(search_term,replace_term,contents,flags=flags)
    self.parent().textCursor().insertText(contents)
  
  def replaceAll(self):
    self.parent().selectAll()
    self.replace(autofind = False)
  
  def find(self, back = False):
    flags = 0x0000
    if self.reCheck.isChecked():
      search_term = QRegExp(self.searchInput.text())
    else:
      search_term = self.searchInput.text()
    if self.matchCase.isChecked():
      flags |= QTextDocument.FindCaseSensitively
    elif self.reCheck.isChecked():
        search_term.setCaseSensitivity(Qt.CaseInsensitive)
    if self.wholeWords.isChecked():
      flags |= QTextDocument.FindWholeWords
    if back:
      flags |= QTextDocument.FindBackward
    start_here = self.parent().textCursor()
    if flags:
      cursor = self.parent().document().find(search_term, start_here, options = flags)
    else:
      cursor = self.parent().document().find(search_term, start_here)
    self.parent().setTextCursor(cursor)
